import os

from requests_futures.sessions import FuturesSession

from savvihub.api.file_object import FileObject
from savvihub.api.savvihub import SavviHubClient
from savvihub.api.types import SavviHubFileObject
from savvihub.common.utils import wait_all_futures, calculate_crc32c


def default_hooks():
    def fn(resp, **kwargs):
        resp.raise_for_status()
    return {
        'response': fn,
    }


class Downloader:
    @classmethod
    def get_files_to_download(cls, context, volume_id, *, snapshot='latest', prefix=''):
        client = SavviHubClient(token=context.token)
        files = client.volume_file_list(volume_id, prefix=prefix, snapshot=snapshot, raise_error=True)
        return files

    @classmethod
    def parallel_download(cls, context, path, files, progressable=None):
        if len(files) <= 0:
            return
        session = FuturesSession(max_workers=context.parallel)
        futures = []
        for file in files:
            file_object = FileObject(file, path)
            if file.path.endswith('/'):
                continue

            progress_callback = None
            if progressable:
                progress = progressable(length=file_object.size, label=file_object.path)
                progress_callback = lambda data: progress.update(len(data))

            future = session.get(
                file.download_url.url,
                stream=True,
                hooks=file_object.download_hooks(callback=progress_callback)
            )
            futures.append(future)
        wait_all_futures(futures)


class Uploader:
    @classmethod
    def get_files_to_upload(cls, base_path, hashmap=None):
        results = []
        for root, dirs, files in os.walk(base_path):
            for name in files:
                name = os.path.join(os.path.abspath(root), name)
                name = name[len(base_path) + 1:] if name.startswith(base_path) else name
                if hashmap and hashmap[name] == calculate_crc32c(os.path.join(base_path, name)):
                    continue
                results.append(name)
        return results

    @classmethod
    def get_hashmap(cls, base_path):
        files = cls.get_files_to_upload(base_path)
        hashmap = dict()
        for file in files:
            path = os.path.join(base_path, file)
            hashmap[file] = calculate_crc32c(path)
        return hashmap

    @classmethod
    def parallel_upload(cls, context, base_path, files, volume_id, *, progressable=None):
        if len(files) <= 0:
            return

        session = FuturesSession(max_workers=context.parallel)
        client = SavviHubClient(token=context.token, session=session)

        futures = []
        for file in files:
            future = client.volume_file_create(
                volume_id,
                file,
                is_dir=False,
                hooks=default_hooks()
            )
            futures.append(future)
        resps = wait_all_futures(futures)

        futures = []
        for i, resp in enumerate(resps):
            file_object = FileObject(SavviHubFileObject(resp.json()), base_path)

            progress_callback = None
            if progressable:
                progress = progressable(length=os.path.getsize(file_object.full_path), label=file_object.path)
                progress_callback = lambda data: progress.update(len(data))

            future = session.put(
                file_object.upload_url.url,
                data=file_object.upload_chunks(callback=progress_callback),
                headers={
                    'content-type': 'application/octet-stream',
                },
                hooks=file_object.upload_hooks(),
            )
            futures.append(future)
        wait_all_futures(futures)

        futures = []
        for file in files:
            future = client.volume_file_uploaded(
                volume_id,
                file,
            )
            futures.append(future)
        wait_all_futures(futures)



import os
from requests_futures.sessions import FuturesSession

import typer

from savvihub.api.savvihub import SavviHubClient
from savvihub.api.uploader import Uploader, Downloader
from savvihub.common.context import Context

dataset_app = typer.Typer()


def parse_dataset_arg(dataset_arg):
    if "/" not in dataset_arg:
        typer.echo("You should specify dataset with workspace. ex) savvihub/mnist")
        return
    workspace, rest = dataset_arg.split("/")
    if ":" in rest:
        dataset, ref = rest.split(":")
    else:
        dataset = rest
        ref = "latest"
    return workspace, dataset, ref


@dataset_app.callback()
def main():
    return


@dataset_app.command()
def create():
    return


@dataset_app.command()
def push(
    dataset_arg: str = typer.Argument(...),
    root_path_arg: str = typer.Argument("."),
):
    root_path = os.path.abspath(root_path_arg)
    if not os.path.exists(root_path) or not os.path.isdir(root_path):
        typer.echo("Must specify directory as an upload path")
        return

    workspace_slug, dataset_slug, ref = parse_dataset_arg(dataset_arg)
    if ref != "latest":
        typer.echo("Cannot edit dataset snapshot. You should create a new snapshot from latest status.")
        return

    # download file list
    context = Context()
    client = SavviHubClient(token=context.token)
    dataset = client.dataset_read(workspace_slug, dataset_slug)

    files = Uploader.get_files_to_upload(root_path)
    if len(files) > 0:
        Uploader.parallel_upload(context, root_path, files, dataset.main_volume_id, progressable=typer.progressbar)


@dataset_app.command()
def pull(
    dataset_arg: str = typer.Argument(...),
    output_path_arg: str = typer.Argument("."),
):
    output_path = os.path.abspath(output_path_arg)
    if os.path.exists(output_path):
        if not os.path.isdir(output_path):
            typer.echo('Must specify directory as an download path')
            return
        if len(os.listdir(output_path)) > 0:
            typer.echo("Must specify empty directory as an output path")
            return
    else:
        os.mkdir(output_path)

    workspace_slug, dataset_slug, ref = parse_dataset_arg(dataset_arg)
    context = Context()

    client = SavviHubClient(token=context.token)
    if ref == 'latest':
        dataset = client.dataset_read(workspace_slug, dataset_slug, raise_error=True)
        files = client.volume_file_list(dataset.main_volume_id, raise_error=True)
    else:
        files = client.dataset_snapshot_file_list(workspace_slug, dataset_slug, ref, raise_error=True)

    if len(files) > 0:
        Downloader.parallel_download(context, output_path, files, progressable=typer.progressbar)

    typer.echo(f'Downloaded {len(files)} files in {output_path}')


if __name__ == "__main__":
    dataset_app()


import os
import tempfile
import time
from collections import defaultdict
from urllib.parse import urlparse, parse_qs

import typer

from savvihub import Context
from savvihub.api.savvihub import SavviHubClient
from savvihub.api.uploader import Downloader, Uploader

volume_app = typer.Typer()


class PathException(Exception):
    pass


def refine_path(path_arg, raise_if_not_empty=False, raise_if_not_exist=False):
    if path_arg.startswith("savvihub://"):
        return path_arg, True

    path = os.path.abspath(path_arg)
    if os.path.exists(path):
        if not os.path.isdir(path):
            raise PathException(f"Must specify directory: {path_arg}")
        if raise_if_not_empty and len(os.listdir(path)) > 0:
            raise PathException(f"Must specify empty directory: {path_arg}")
    else:
        if raise_if_not_exist:
            raise PathException(f"Must specify directory: {path_arg}")
        os.mkdir(path)

    return path, False


def split_remote_path(remote_path):
    u = urlparse(remote_path)
    volume_id = u.netloc
    prefix = u.path
    query = parse_qs(u.query)
    snapshot = query.get('snapshot', ['latest'])
    if len(snapshot) != 1:
        typer.echo(f'Invalid snapshots: {remote_path}')
    return volume_id, prefix, snapshot[0]


@volume_app.callback()
def main():
    return


@volume_app.command()
def cp(
    source_path_arg: str = typer.Argument(...),
    dest_path_arg: str = typer.Argument(...),
    watch: bool = typer.Option(False, "-w", "--watch"),
):
    try:
        source_path, is_source_remote = refine_path(source_path_arg, raise_if_not_exist=True)
        dest_path, is_dest_remote = refine_path(dest_path_arg, raise_if_not_empty=True)
    except PathException as e:
        typer.echo(str(e))
        return

    context = Context()
    client = SavviHubClient(token=context.token)

    hashmap = defaultdict(lambda: "")

    while True:
        if is_source_remote and is_dest_remote:
            # remote -> remote
            source_volume_id, source_prefix, source_snapshot_ref = split_remote_path(source_path)
            dest_volume_id, dest_prefix, dest_snapshot_ref = split_remote_path(dest_path)
            if dest_snapshot_ref != 'latest':
                typer.echo(f'Cannot write to snapshots: {dest_path}')
                return
            files = Downloader.get_files_to_download(context, source_volume_id, snapshot=source_snapshot_ref)
            files = [file for file in files if hashmap[file.path] != file.hash]

            if len(files) > 0:
                tmp_dir = tempfile.mkdtemp()
                Downloader.parallel_download(context, tmp_dir, files, progressable=typer.progressbar)
                Uploader.parallel_upload(context, tmp_dir, files, dest_volume_id, progressable=typer.progressbar)

                for file in files:
                    hashmap[file.path] = file.hash

        elif is_source_remote and not is_dest_remote:
            # remote -> local
            source_volume_id, source_prefix, source_snapshot_ref = split_remote_path(source_path)
            files = Downloader.get_files_to_download(context, source_volume_id, prefix=source_prefix, snapshot=source_snapshot_ref)
            files = [file for file in files if hashmap[file.path] != file.hash]

            typer.echo(f'Find {len(files)} files to download.')
            if len(files) > 0:
                Downloader.parallel_download(context, dest_path, files, progressable=typer.progressbar)

                for file in files:
                    hashmap[file.path] = file.hash

        elif not is_source_remote and is_dest_remote:
            # local -> remote
            dest_volume_id, dest_prefix, dest_snapshot_ref = split_remote_path(dest_path)
            if dest_snapshot_ref != 'latest':
                typer.echo(f'Cannot write to snapshots: {dest_path}')
                return

            files = Uploader.get_files_to_upload(source_path, hashmap)
            files = [f'{os.path.join(dest_prefix, file)}' for file in files]

            typer.echo(f'Find {len(files)} files to upload.')
            if len(files) > 0:
                Uploader.parallel_upload(context, source_path, files, dest_volume_id, progressable=typer.progressbar)

                hashmap = Uploader.get_hashmap(source_path)
        else:
            typer.echo('Cannot copy volume from local to local')
            return

        if not watch:
            return

        time.sleep(10)



import os
import time
from collections import defaultdict

import typer

from savvihub.api.savvihub import SavviHubClient
from savvihub.api.uploader import Downloader, Uploader
from savvihub.common.context import Context


artifact_app = typer.Typer()


@artifact_app.callback()
def main():
    return


@artifact_app.command()
def upload(
    watch: bool = typer.Option(False, "-w", "--watch"),
    output_path_arg: str = typer.Argument("."),
):
    output_path = os.path.abspath(output_path_arg)
    if not os.path.exists(output_path) or not os.path.isdir(output_path):
        typer.echo("Must specify directory as an upload path")
        return

    context = Context(experiment_required=True)
    experiment = context.experiment

    hashmap = None
    while True:
        files = Uploader.get_files_to_upload(output_path, hashmap)

        typer.echo(f'Find {len(files)} files to upload.')
        if len(files) > 0:
            Uploader.parallel_upload(context, output_path, files, experiment.output_volume_id, progressable=typer.progressbar)
            hashmap = Uploader.get_hashmap(output_path)

        if not watch:
            return

        time.sleep(10)


@artifact_app.command()
def download(
    prefix: str = typer.Option("", "--prefix"),
    watch: bool = typer.Option(False, "-w", "--watch"),
    output_path_arg: str = typer.Argument("."),
):
    output_path = os.path.abspath(output_path_arg)
    if os.path.exists(output_path):
        if not os.path.isdir(output_path):
            typer.echo("Must specify directory as an download path")
            return
        if len(os.listdir(output_path)) > 0:
            typer.echo("Must specify empty directory as an output path")
            return
    else:
        os.mkdir(output_path)

    context = Context(experiment_required=True)
    experiment = context.experiment

    hashmap = defaultdict(lambda: "")

    while True:
        files = Downloader.get_files_to_download(context, experiment.output_volume_id, prefix=prefix)
        files = [file for file in files if hashmap[file.path] != file.hash]

        typer.echo(f'{len(files)} files to downloaded')
        if len(files) > 0:
            Downloader.parallel_download(context, output_path, files, progressable=typer.progressbar)
            for file in files:
                hashmap[file.path] = file.hash

        if not watch:
            return

        time.sleep(10)



import inspect
import typeguard
import typing


class AnnotatedObject:
    def __init__(self, d: dict):
        resolved_types = {}
        for cls in self.__class__.mro():
            annotations = getattr(cls, '__annotations__', None)
            if annotations is None:
                continue

            for k, type_ in annotations.items():
                if k in resolved_types:
                    continue

                v = d.get(k, None)
                resolved_types[k] = type_

                if inspect.isclass(type_) and issubclass(type_, AnnotatedObject):
                    v = type_(v)
                else:
                    typeguard.check_type(k, v, type_)
                setattr(self, k, v)
        self.d = d

    def __str__(self):
        return f'[{self.__class__.__name__}] {self.d}'

    @property
    def dict(self):
        return self.d


class SavviHubUser(AnnotatedObject):
    id: int
    username: str
    name: str
    github_authorized: bool


class SavviHubWorkspace(AnnotatedObject):
    id: int
    slug: str


class SavviHubFileObject(AnnotatedObject):
    class FileActionURL(AnnotatedObject):
        url: str

    path: str

    size: int
    hash: str

    download_url: FileActionURL
    upload_url: FileActionURL


class SavviHubKernelImage(AnnotatedObject):
    id: int
    image_url: str
    name: str


class SavviHubKernelResource(AnnotatedObject):
    id: int
    name: str
    cpu_limit: float
    mem_limit: str


class SavviHubDataset(AnnotatedObject):
    id: int
    workspace: SavviHubWorkspace
    slug: str
    main_volume_id: int


class SavviHubSnapshot(AnnotatedObject):
    id: int
    ref: str
    name: str
    size: int


class SavviHubProject(AnnotatedObject):
    id: int


class SavviHubExperiment(AnnotatedObject):
    id: int
    number: int
    status: str
    message: str
    output_volume_id: int


class SavviHubExperimentLog(AnnotatedObject):
    name: str
    stream: str
    message: str
    timestamp: float
    checksum: str


class SavviHubListResponse(AnnotatedObject):
    results: typing.List


class PaginatedMixin(AnnotatedObject):
    total: int
    startCursor: typing.Optional[str]
    endCursor: typing.Optional[str]
    results: typing.List



import re
from datetime import datetime
from typing import List

import typer
from terminaltables import AsciiTable
import inquirer

from savvihub.common.context import Context
from savvihub.common.git import GitRepository, InvalidGitRepository
from savvihub.savvihub_cli.inquirer import get_choices, parse_title
from savvihub.api.savvihub import SavviHubClient
from savvihub.common.constants import CUR_DIR, INQUIRER_NAME_IMAGE, INQUIRER_NAME_RESOURCE, \
    INQUIRER_NAME_COMMAND, WEB_HOST, INQUIRER_NAME_DATASET, INQUIRER_NAME_DATASET_REF, \
    INQUIRER_NAME_DATASET_MOUNT_PATH, DEFAULT_SAVVIHUBFILE_YAML
from savvihub.savvihub_cli.errors import get_error_message
from savvihub.common.utils import *

experiment_app = typer.Typer()


@experiment_app.callback()
def main():
    """
    Perform your experiment with Savvihub
    """
    return


@experiment_app.command()
def list():
    """
    List of experiments
    """
    try:
        git_repo = GitRepository()
    except InvalidGitRepository as e:
        typer.echo(str(e))
        return

    context = Context(project_required=True, project_config_path=git_repo.get_savvihub_config_file_path())
    client = SavviHubClient(token=context.token)

    experiments = client.experiment_list(context.workspace, context.project, raise_error=True)
    table = AsciiTable([
        ['Number', 'Status', 'Message'],
        *[[e.number, e.status, e.message] for e in experiments],
    ])
    typer.echo(table.table)


@experiment_app.command()
def logs(
    experiment_number: int = typer.Argument(...),
):
    """
    View experiment logs
    """
    try:
        git_repo = GitRepository()
    except InvalidGitRepository as e:
        typer.echo(str(e))
        return

    context = Context(project_required=True, project_config_path=git_repo.get_savvihub_config_file_path())
    client = SavviHubClient(token=context.token)

    logs = client.experiment_log(context.workspace, context.project, experiment_number, raise_error=True)
    for log in logs:
        print(datetime.fromtimestamp(log.timestamp), log.message)


@experiment_app.command()
def run(
    command_arg: str = typer.Option(None, "-c"),
    image_arg: str = typer.Option(None, "-i"),
    resource_arg: str = typer.Option(None, "-r"),
    env_vars_arg: List[str] = typer.Option([], "-e"),
    dataset_mount_args: List[str] = typer.Option([], "--dataset"),
):
    """
    Run an experiment in Savvihub
    """
    try:
        git_repo = GitRepository()
    except InvalidGitRepository as e:
        typer.echo(str(e))
        return

    context = Context(project_required=True, project_config_path=git_repo.get_savvihub_config_file_path())
    client = SavviHubClient(token=context.token)

    def find_from_args(options, selector):
        for option in options:
            if selector(option):
                return option
        return None

    selected_image, selected_resource, start_command = None, None, None

    images = client.kernel_image_list(context.workspace)
    if image_arg:
        selected_image = find_from_args(images, lambda x: x.image_url == image_arg.strip())
        if selected_image is None:
            typer.echo(f'Cannot find image {image_arg}.')
            return
    else:
        answers = inquirer.prompt([inquirer.List(
            INQUIRER_NAME_IMAGE,
            message="Please choose a kernel image",
            choices=[f'[{i+1}] {image.image_url} ({image.name})' for i, image in enumerate(images)],
        )])
        answer = int(re.findall(r"[\d+]", answers.get(INQUIRER_NAME_IMAGE))[0]) - 1
        selected_image = images[answer]

    resources = client.kernel_resource_list(context.workspace)
    if resource_arg:
        selected_resource = find_from_args(resources, lambda x: x.name == resource_arg.strip())
        if selected_resource is None:
            typer.echo(f'Cannot find resource {resource_arg}.')
            return
    else:
        answers = inquirer.prompt([inquirer.List(
            INQUIRER_NAME_RESOURCE,
            message="Please choose a kernel resource",
            choices=[f'[{i+1}] {resource.name}' for i, resource in enumerate(resources)]
        )])
        answer = int(re.findall(f"[\d+]", answers.get(INQUIRER_NAME_RESOURCE))[0]) - 1
        selected_resource = resources[answer]

    if command_arg:
        start_command = command_arg
    else:
        answers = inquirer.prompt([inquirer.Text(
            INQUIRER_NAME_COMMAND,
            message="Start command",
            default="python main.py",
        )])
        answer = answers.get(INQUIRER_NAME_COMMAND)
        start_command = answer

    dataset_mounts_parsed = []
    for dataset_mount in dataset_mount_args:
        # parse dataset and volume
        if ':' not in dataset_mount:
            typer.echo(f'Invalid dataset slug: {dataset_mount}. '
                       f'You should specify dataset mount location.\n'
                       f'ex) savvihub/mnist@3d1e0f:/input/dataset1')
            return
        dataset, mount_path = dataset_mount.split(':')

        # parse snapshot ref
        if '@' in dataset:
            dataset_slug, snapshot_ref = dataset.split('@')
        else:
            dataset_slug = dataset
            snapshot_ref = 'latest'

        if '/' not in dataset_slug:
            typer.echo(f'Invalid dataset slug: {dataset_mount}. '
                       f'You should specify dataset with workspace.\n'
                       f'ex) savvihub/mnist@3d1e0f:/input/dataset1')
            return
        workspace, dataset = dataset_slug.split('/')

        # read dataset or snapshot
        dataset_obj = client.dataset_read(workspace, dataset)
        if not dataset_obj:
            typer.echo(f'Invalid dataset: {dataset_mount}. '
                       f'Please check your dataset exist in savvihub.\n')
            return
        if snapshot_ref != 'latest':
            snapshot_obj = client.snapshot_read(dataset_obj.main_volume_id, snapshot_ref)
            if not snapshot_obj:
                typer.echo(f'Invalid dataset snapshots: {dataset_mount}. '
                           f'Please check your dataset and snapshot exist in savvihub.')
                return

        dataset_mounts_parsed.append(dict(
            dataset_id=dataset_obj.id,
            snapshot_ref=snapshot_ref,
            mount_path=mount_path,
        ))

    revision_or_branch, is_revision = git_repo.get_remote_revision_or_branch()
    revision, branch = None, None
    if is_revision:
        revision = revision_or_branch
    else:
        branch = revision_or_branch

    diff_file = git_repo.get_current_diff_file(revision_or_branch)

    res = client.experiment_create(
        workspace=context.workspace,
        project=context.project,
        image_id=selected_image.id,
        resource_spec_id=selected_resource.id,
        git_ref=revision_or_branch,
        start_command=start_command,
        dataset_mount_infos=dataset_mounts_parsed,
    )

    diff_file.close()

    res_data = res.json()
    if res.status_code == 400:
        typer.echo(get_error_message(res_data))
        return
    res.raise_for_status()

    experiment_number = res_data.get('number')
    typer.echo(f"Experiment {experiment_number} is running. Check the experiment status at below link")
    typer.echo(f"{WEB_HOST}/{context.workspace}/{context.project}/"
               f"experiments/{experiment_number}")
    return


if __name__ == "__main__":
    experiment_app()



import os
import requests
import yaml

from savvihub.api.types import SavviHubUser
from savvihub.common.constants import DEFAULT_CONFIG_PATH, CUR_DIR, DEFAULT_SAVVIHUBFILE_YAML
from savvihub.common.experiment import Experiment


class Value:
    def __init__(self, global_config_yml=None, savvihub_file_yml=None, env=None, computed=None, default_value=None):
        self.global_config_yml = global_config_yml
        self.savvihub_file_yml = savvihub_file_yml
        self.env = env
        self.computed = computed
        self.default_value = default_value


def get_experiment(token=None, **kwargs):
    if not token:
        return

    from savvihub.api.savvihub import SavviHubClient
    client = SavviHubClient(token=token)

    try:
        savvihub_experiment = client.experiment_read(raise_error=True)
    except requests.exceptions.HTTPError as e:
        print(f"Http Error: {e}")
        return
    except requests.exceptions.ConnectionError as e:
        print(f"Connection Error:{e}")
        return
    except requests.exceptions.Timeout as e:
        print(f"Timeout Error: {e}")
        return
    except requests.exceptions.TooManyRedirects as e:
        print(f"Too many Redirects Error: {e}")
        return
    except requests.exceptions.RequestException as e:
        raise SystemExit(e)

    from savvihub.common.experiment import Experiment
    return Experiment.from_given(savvihub_experiment, client)


def get_my_info(token=None, **kwargs):
    from savvihub.api.savvihub import SavviHubClient
    client = SavviHubClient(token=token)
    return client.get_my_info()


class Context:
    token = Value(global_config_yml="token", env="EXPERIMENT_ACCESS_TOKEN", default_value=None)
    parallel = Value(env="PARALLEL", default_value=20)
    workspace = Value(savvihub_file_yml="workspace")
    project = Value(savvihub_file_yml="project")

    me: SavviHubUser = Value(computed=get_my_info)
    experiment: Experiment = Value(computed=get_experiment)

    def __init__(self, login_required=False, experiment_required=False, project_required=False, project_config_path=None):
        if project_required and not project_config_path:
            raise Exception('Project is required but project_config_path is not given.')

        attrs = dict()
        for key in dir(self):
            attr = getattr(self, key)
            if isinstance(attr, Value):
                attrs[key] = attr

        data = dict()

        global_config = self.load_global_config_file()

        savvihub_yml_config = self.load_experiment_config_file(project_config_path)

        for key, attr in attrs.items():
            # Set default value
            data[key] = attr.default_value

            # Load from file
            if attr.global_config_yml and global_config:
                value = global_config.get(attr.global_config_yml, None)
                if value is not None:
                    data[key] = value

            if attr.savvihub_file_yml and savvihub_yml_config:
                value = savvihub_yml_config.get(attr.savvihub_file_yml, None)
                if value is not None:
                    data[key] = value

            # Load from env
            if attr.env:
                value = os.environ.get(attr.env, None)
                if value is not None:
                    data[key] = value

        # calculate computed
        for key, attr in attrs.items():
            if attr.computed:
                value = attr.computed(**data)
                if value:
                    data[key] = value

        for k, v in data.items():
            setattr(self, k, v)

        if login_required and self.me is None:
            if self.token:
                raise Exception('Login required. You should call `savvi init` first.')
            else:
                raise Exception('Token expired. You should call `savvi init` first.')

        if experiment_required and self.experiment is None:
            raise Exception('Experiment token required.')

    @staticmethod
    def load_global_config_file():
        try:
            with open(DEFAULT_CONFIG_PATH, "r") as f:
                global_config = yaml.full_load(f)
                if global_config:
                    return global_config
        except FileNotFoundError:
            return

    @staticmethod
    def load_experiment_config_file(experiment_yml):
        try:
            if not experiment_yml:
                return
            with open(experiment_yml, "r") as f:
                experiment_config = yaml.full_load(f)
                if experiment_config:
                    return experiment_config
        except FileNotFoundError:
            return



import os

from savvihub.api.types import SavviHubFileObject
from savvihub.common.utils import read_in_chunks


class FileObject(SavviHubFileObject):
    def __init__(self, file_object: SavviHubFileObject, base_path):
        super().__init__(file_object.dict)
        self.base_path = base_path
        self.full_path = os.path.join(base_path, self.path)

    def upload_chunks(self, *, callback=None):
        return read_in_chunks(self.full_path, callback=callback)

    def upload_hooks(self, *, log=None):
        def fn(resp, **kwargs):
            resp.raise_for_status()
        return {
            'response': fn,
        }

    def download_hooks(self, *, callback=None):
        def fn(resp, **kwargs):
            os.makedirs(os.path.dirname(self.full_path), exist_ok=True)
            with open(self.full_path, 'wb') as f:
                for chunk in resp.iter_content(chunk_size=8192):
                    f.write(chunk)
                    if callback:
                        callback(chunk)
        return {
            'response': fn,
        }



import os

import typer

from savvihub import Context
from savvihub.api.savvihub import SavviHubClient
from savvihub.common.constants import WEB_HOST
from savvihub.common.git import GitRepository, InvalidGitRepository
from savvihub.common.utils import make_file, remove_file
from savvihub.savvihub_cli.yml_loader import ProjectYmlLoader

project_app = typer.Typer()


@project_app.callback()
def main():
    return


@project_app.command()
def init(
    slug: str = typer.Argument(..., help="Type workspace/project as an argument"),
):
    """
    Initialize a new experiment configuration file with workspace/project
    """
    try:
        git_repo = GitRepository()
    except InvalidGitRepository as e:
        typer.echo(str(e))
        return

    config_file_path = git_repo.get_savvihub_config_file_path()
    if os.path.exists(config_file_path):
        answer = typer.prompt(f'Savvihub config file already exists in {config_file_path} Overwrite file? [Y/n] ')
        if answer.lower().startswith('y'):
            remove_file(config_file_path)
        else:
            return

    workspace_slug, project_slug = slug.split("/")
    context = Context(login_required=True)
    client = SavviHubClient(token=context.token)

    workspace = client.workspace_read(workspace_slug)
    if workspace is None:
        typer.echo(f'Cannot find workspace {workspace}.')
        return

    project = client.project_read(workspace_slug, project_slug)
    if project is None:
        owner, repo = git_repo.get_github_repo()
        create_new_project = typer.prompt(f'Project not found. Create new project in SavviHub? [Y/n] ')
        if create_new_project.lower().strip().startswith('y'):
            if not context.me.github_authorized:
                typer.echo(f'You should authorize github first.\nhttp://{WEB_HOST}/github/authorize/')
                return

            typer.echo(f'Workspace: {workspace_slug}\nProject: {project_slug}\nGit: github.com/{owner}/{repo}')
            client.project_github_create(workspace_slug, project_slug, owner, repo, raise_error=True)
            typer.echo(f'Project created successfully.\nhttp://{WEB_HOST}/{workspace_slug}/{project_slug}')

    data = {
        'workspace': workspace_slug,
        'project': project_slug,
    }

    make_file(config_file_path)
    yml_loader = ProjectYmlLoader(config_file_path)
    yml_loader.write(data)

    typer.echo(f"Experiment config successfully made in {config_file_path}")
